﻿Imports System
Imports System.Collections.Generic
Imports System.ComponentModel
Imports System.Data
Imports System.Drawing
Imports System.Linq
Imports System.Text
Imports System.Threading.Tasks
Imports System.Windows.Forms
Imports FarPoint.Win.Spread
Imports GrapeCity
Imports GrapeCity.Core
Imports GrapeCity.CalcEngine
Imports GrapeCity.Spreadsheet
Imports Microsoft.VisualBasic

Namespace SpreadNETZetaCalculator
    ''' <summary>
    ''' Spread.NET 12 Windows Forms ZETA Function Calcultor / Non-trivial Root Finder 
    ''' </summary>
    Partial Public Class ZetaCalculatorForm
        Inherits Form
        ''' <summary>
        ''' Create the ZetaCalculatorForm
        ''' </summary>
        Public Sub New()
            InitializeComponent()
            fpSpread1.AddCustomFunction(New ZetaFunction())
            formulaTextBox1.Attach(fpSpread1)
            fpSpread1.StatusBarVisible = True
        End Sub
        ''' <summary>
        ''' Returns the input values for the Zeta calculation.
        ''' </summary>
        ''' <param name="initval">The starting value to use for the imaginary component (the real component used is always 0.5)</param>
        ''' <param name="incr">The increment to use for calculating successive values</param>
        ''' <param name="rc">The number of values of Zeta to calculate</param>
        ''' <param name="iter">The number of iterations to execute the calculation -- higher values will result in greater precision but take more time</param>
        ''' <returns>True if the input values are valid and successfully retrieved, False otherwise</returns>
        Private Function GetInputValues(ByRef initval As Double, ByRef incr As Double, ByRef rc As Integer, ByRef iter As Long) As Boolean
            If Not Double.TryParse(textBoxStart.Text, initval) Then
                MessageBox.Show("Please enter a double value for Start Value.")
                textBoxStart.Focus()
                Return False
            End If
            If Not Double.TryParse(textBoxIncrement.Text, incr) Then
                MessageBox.Show("Please enter a double value for Increment.")
                textBoxIncrement.Focus()
                Return False
            End If
            If Not Integer.TryParse(textBoxCount.Text, rc) Then
                MessageBox.Show("Please enter an integer value for Count.")
                textBoxCount.Focus()
                Return False
            End If
            Long.TryParse(textBoxIter.Text, iter)
            Return True
        End Function
        ''' <summary>
        ''' Returns the number format speifying the number of decimal places used for increment.
        ''' </summary>
        ''' <returns>
        ''' The number format specifying the number of decimal places used for increment.
        ''' </returns>
        ''' <remarks>
        ''' The calculator formats the first column using the number of decimal places for increment.
        ''' </remarks>
        Private Function GetIncrementNumberFormat() As String
            Dim incr As Double = 0.1
            If Not Double.TryParse(textBoxIncrement.Text, incr) Then Return "0.0"
            Dim sb As New StringBuilder
            sb.Append("0.")
            While incr < 1
                sb.Append("0")
                incr *= 10
            End While
            Return sb.ToString()
        End Function
        ''' <summary>
        ''' Handle the "Calculate" button click.
        ''' </summary>
        ''' <param name="sender"></param>
        ''' <param name="e"></param>
        ''' <remarks>
        ''' This event handler reinitialized the worksheet with the newly-generated Zeta calculation table.
        ''' This code sets the formula to calculate ZETA as an array formula, and the spreadsheet calculates the result
        ''' once, and sets the result to the range all at once using an array.
        ''' The actual calculation is handled inside the custom function that is active (ZetaFunction or ZetaFunctionInfo).
        ''' The title bar text is updated at the end with timing and call count information about the custom function execution.
        ''' </remarks>
        Private Sub buttonCalculate_Click(ByVal sender As Object, ByVal e As EventArgs) Handles buttonCalculate.Click
            Dim initval As Double = 0.0
            Dim incr As Double = 0.1
            Dim rc As Integer = 1000
            Dim iter As Long = 10000
            If Not GetInputValues(initval, incr, rc, iter) Then Return
            fpSpread1.SuspendLayout()
            Cursor = Cursors.WaitCursor
            Text = "Calculating, please wait..."
            ' inialize sheet with Zeta calculation table headers
            Dim sheet As SheetView = fpSpread1.ActiveSheet
            sheet.Reset()
            sheet.RowCount = rc + 1
            sheet.ColumnCount = 5
            sheet.ReferenceStyle = FarPoint.Win.Spread.Model.ReferenceStyle.R1C1
            sheet.Cells(0, 0).Text = "bi"
            sheet.Cells(0, 1).Text = "Re( ζ(0.5+bi) )"
            sheet.Cells(0, 2).Text = "Im( ζ(0.5+bi) )"
            sheet.Cells(0, 3).Text = "ζ(0.5+bi)"
            sheet.Cells(0, 4).Text = "IMABS( ζ(0.5+bi) )"
            sheet.Cells(0, 0, 0, 4).HorizontalAlignment = CellHorizontalAlignment.Center
            ' apply built-in styles for table header cells
            Dim workbook As IWorkbook = fpSpread1.AsWorkbook()
            workbook.ActiveSheet.Cells(0, 0, 0, 4).ApplyStyle(BuiltInStyle.Percent20Accent3)
            workbook.ActiveSheet.Cells(0, 0, 0, 4).ApplyStyle(BuiltInStyle.Heading3)
            ' initialize the values to be calculated
            For r As Integer = 1 To rc
                sheet.Cells(r, 0).Value = initval + incr * r
            Next
            ' format the cells to show the number decimal places appropriate for increment
            fpSpread1.AsWorkbook().Worksheets(0).Cells(1, 0, rc, 0).NumberFormat = GetIncrementNumberFormat()
            ' create the formula to calculate the ZETA function along the critical line Re = 0.5 for the 
            ' set of points starting with the initial value and using the specified increment and iterations
            Dim formula As String = "ZETA(0.5,RC1," + rc.ToString().Trim() + "," + incr.ToString().Trim() + "," + iter.ToString().Trim() + ")"
            sheet.Cells(1, 1, rc, 2).FormulaArray = formula
            ' apply the COMPLEX function to the two values returned for the real (in column 1) and imaginary (in column 2) parts to get the resulting complex number calculated
            sheet.Cells(1, 3, rc, 3).Formula = "COMPLEX(RC2,RC3)"
            ' apply IMABS function to resulting complex number -- this is equivalent to SQRT(RC2*RC2+RC3*RC3) and computes the length of the complex vector
            ' IMABS values less than increment are considered "close" to a root and highlighted if the value is a local minimum
            sheet.Cells(1, 4, rc, 4).Formula = "IMABS(RC4)"
            ' set column widths
            sheet.SetColumnWidth(0, CInt(sheet.GetPreferredColumnWidth(0)))
            sheet.SetColumnWidth(1, 150)
            sheet.SetColumnWidth(2, 150)
            sheet.SetColumnWidth(3, CInt(sheet.GetPreferredColumnWidth(3)))
            sheet.SetColumnWidth(4, CInt(sheet.GetPreferredColumnWidth(4)))
            ' enable auto filter on the range
            workbook.ActiveSheet.Cells(0, 0, rc, 4).AutoFilter()
            ' freeze first row from scrolling to keep table headers visible
            sheet.FrozenRowCount = 1
            ' renumber row headings 
            sheet.StartingRowNumber = 0
            ' clear the text in the first row header
            sheet.RowHeader.Cells(0, 0).Text = " "
            ' map the clipboard shortcut keys to special actions that copy only visible values as text
            ' (so the sheet can be filtered and copy/paste works as expected)
            fpSpread1.GetInputMap(InputMapMode.WhenFocused).Put(New Keystroke(Keys.C, Keys.Control), SpreadActions.ClipboardCopyAsStringSkipHidden)
            fpSpread1.GetInputMap(InputMapMode.WhenFocused).Put(New Keystroke(Keys.Insert, Keys.Control), SpreadActions.ClipboardCopyAsStringSkipHidden)
            ' create a conditional format to highlight the root (local minima IMABS < 1) values with yellow
            Dim conditionformula As String = "AND(RC5<" + incr.ToString() + ",RC5<R[1]C5,RC5<R[-1]C5)"
            Dim formulaRule As FormulaConditionalFormattingRule = New FormulaConditionalFormattingRule(conditionformula)
            formulaRule.BackColor = System.Drawing.Color.Yellow
            sheet.SetConditionalFormatting(1, 0, rc, 5, New IConditionalFormattingRule() {formulaRule})
            ' cleanup, update window text with function call count and timing stats, enable filter roots button
            fpSpread1.ResumeLayout()
            Cursor = Cursors.[Default]
            If checkBoxFunctionInfo.Checked Then
                Text = "Zeta Calculation Complete! [ZetaFunctionInfo Call #" + ZetaFunctionInfo.cnt.ToString() + " in " & (ZetaFunctionInfo.t2 - ZetaFunctionInfo.t1).ToString() & "]"
            Else
                Text = "Zeta Calculation Complete! [ZetaFunction Call #" + ZetaFunction.cnt.ToString() + " in " & (ZetaFunction.t2 - ZetaFunction.t1).ToString() & "]"
            End If
            buttonFilterRootsOnly.Enabled = True
        End Sub
        ''' <summary>
        ''' Apply a filter on the range to show only the root values.
        ''' Creates a color filter for the first column in the worksheet.
        ''' </summary>
        ''' <param name="sender"></param>
        ''' <param name="e"></param>
        Private Sub buttonFilterRootsOnly_Click(ByVal sender As Object, ByVal e As EventArgs) Handles buttonFilterRootsOnly.Click
            Dim workbook As IWorkbook = fpSpread1.AsWorkbook()
            Dim worksheet As IWorksheet = workbook.ActiveSheet
            Dim fill As Fill = New Fill(GrapeCity.Spreadsheet.Color.FromKnownColor(GrapeCity.Core.KnownColor.Yellow))
            worksheet.Cells(0, 0, worksheet.RowCount - 1, 4).AutoFilter(0, fill)
        End Sub
        ''' <summary>
        ''' Toggles between using the new-style calculation function ZetaFunction and the old-style calculation function ZetaFunctionInfo.
        ''' The new-style calculation function is recommended as it is slightly faster and easier to use than the old-style calculation function,
        ''' which must be wrapped internally using GrapeCity.CalcEngine APIs to work in the new GrapeCity.CalcEngine.
        ''' </summary>
        ''' <param name="sender"></param>
        ''' <param name="e"></param>
        Private Sub checkBoxFunctionInfo_CheckedChanged(ByVal sender As Object, ByVal e As EventArgs) Handles checkBoxFunctionInfo.CheckedChanged
            If checkBoxFunctionInfo.Checked Then
                ' this method is marked obsolete and deprecated -- you should port your old FunctionInfo class to a Function class
                fpSpread1.ActiveSheet.AddCustomFunction(New ZetaFunctionInfo())
            Else
                fpSpread1.AddCustomFunction(New ZetaFunction())
            End If
        End Sub
        ''' <summary>
        ''' Event handler for the LeaveCell event.
        ''' </summary>
        ''' <param name="sender"></param>
        ''' <param name="e"></param>
        ''' <remarks>
        ''' This event handler enables and disables the GoalSeek button based on whether the active row is a highlighted root value.
        ''' </remarks>
        Private Sub fpSpread1_LeaveCell(ByVal sender As Object, ByVal e As LeaveCellEventArgs) Handles fpSpread1.LeaveCell
            Dim workbook As IWorkbook = fpSpread1.AsWorkbook()
            ' get the interior color of the cell becoming active, enable goal seek button if it's a highlighted root value
            Dim c As GrapeCity.Spreadsheet.Color = workbook.ActiveSheet.Cells(e.NewRow, e.NewColumn).DisplayFormat.Interior.Color
            If c.Equals(GrapeCity.Spreadsheet.Color.FromKnownColor(GrapeCity.Core.KnownColor.Yellow)) Then
                buttonGoalSeek.Enabled = True
            Else
                buttonGoalSeek.Enabled = False
            End If
        End Sub
        ''' <summary>
        ''' Do a high-precision goal seek to 4 decimal places
        ''' </summary>
        ''' <param name="sender"></param>
        ''' <param name="e"></param>
        ''' <remarks>
        ''' The spreadsheet has built-in GoalSeek logic, but that logic is limited to 2 decimal places.
        ''' This function uses the CalculationEngine to perform a goal seek to 4 decimal places.
        ''' </remarks>
        Private Sub buttonGoalSeek_Click(ByVal sender As Object, ByVal e As EventArgs) Handles buttonGoalSeek.Click
            Dim t2 As DateTime, t1 As DateTime = DateTime.Now ' for timing the goal-seek
            Dim wb As IWorkbook = fpSpread1.AsWorkbook()
            Dim calcengine As CalculationEngine = wb.WorkbookSet.CalculationEngine
            Dim ws As IWorksheet = wb.ActiveSheet
            Dim sheet As SheetView = fpSpread1.ActiveSheet
            Dim initval As Double = 0.0
            Dim incr As Double = 0.1
            Dim rc As Integer = 1000
            Dim iter As Long = 10000
            Dim row As Integer = sheet.ActiveRowIndex
            Dim cnt As Integer ' for counting the number of ZETA function calls performed during goal-seek
            Dim cnt2 As Integer
            Dim maxchange As Double ' current decimal place of search
            Dim formula, formulaNext As String ' formula for value and formula for next value to be tested
            Dim result, resultNext As IValue ' calculated result of formula for value and formula for next value
            ' temp variables for calculating IMABS of result and resultNext
            Dim temp, temp1, temp2, tempNext, tempNext1, tempNext2 As Double
            Dim nextDecimalPlace As Boolean ' flag for moving to next decimal place
            Dim value As Double ' current value being tested
            ' save initial ZETA call count
            If checkBoxFunctionInfo.Checked Then
                cnt = ZetaFunctionInfo.cnt
            Else
                cnt = ZetaFunction.cnt
            End If
            If Not GetInputValues(initval, incr, rc, iter) Then Return
            ' force 100K interations for improved accuracy
            iter = 100000
            ' do goal seek on the cell (row, 0):
            value = Math.Round(CDbl(ws.Cells(row, 0).Value), 1) + 0.04
            ws.Cells(row, 0).NumberFormat = "0.0000"
            Cursor = Cursors.WaitCursor
            maxchange = 0.01 ' start changing the hundredths place, then move on to each sucessive decimal place until maxchange = 1e-10
            Do
                nextDecimalPlace = False ' reset flag
                ' calculate value
                formula = "ZETA(0.5," + value.ToString().Trim() + ",1,0," + iter.ToString().Trim() + ")"
                result = calcengine.Evaluate(formula, ws, New GrapeCity.Spreadsheet.CellReference(row, 1))
                temp1 = result.GetValuesArray()(0, 0).GetNumber()
                temp2 = result.GetValuesArray()(0, 1).GetNumber()
                ' calculate absolute values of results
                temp = Math.Sqrt(temp1 * temp1 + temp2 * temp2)
                Do
                    ' calculate next value (value + maxchange) using CalculationEngine.Evaluate
                    formulaNext = "ZETA(0.5," + (value + maxchange).ToString().Trim() + ",1,0," + iter.ToString().Trim() + ")"
                    resultNext = calcengine.Evaluate(formulaNext, ws, New GrapeCity.Spreadsheet.CellReference(row, 1))
                    tempNext1 = resultNext.GetValuesArray()(0, 0).GetNumber()
                    tempNext2 = resultNext.GetValuesArray()(0, 1).GetNumber()
                    ' calculate absolute values of results
                    tempNext = Math.Sqrt(tempNext1 * tempNext1 + tempNext2 * tempNext2)
                    If tempNext < temp Then
                        value += maxchange ' next value is closer, increment value and continue checking next higher value
                        temp = tempNext
                    Else ' next value is not closer, check previous value(s)
                        Do
                            ' calculate previous value
                            formulaNext = "ZETA(0.5," + (value - maxchange).ToString().Trim() + ",1,0," + iter.ToString().Trim() + ")"
                            resultNext = calcengine.Evaluate(formulaNext, ws, New GrapeCity.Spreadsheet.CellReference(row, 1))
                            tempNext1 = resultNext.GetValuesArray()(0, 0).GetNumber()
                            tempNext2 = resultNext.GetValuesArray()(0, 1).GetNumber()
                            ' calculate absolute value
                            tempNext = Math.Sqrt(tempNext1 * tempNext1 + tempNext2 * tempNext2)
                            If tempNext < temp Then
                                ' previous value is closer, decrement value and continue checking next previous value
                                value -= maxchange
                                temp = tempNext
                            Else
                                ' previous value is not closer, current value is closest -- continue to next decimal place
                                nextDecimalPlace = True
                                ' decrement current decimal place value by 1 to avoid rounding errors, then make the new last decimal place a 4
                                value = value + 4 * (maxchange / 10)
                            End If
                        Loop Until nextDecimalPlace
                    End If
                Loop Until nextDecimalPlace
                maxchange *= 0.1
            Loop While maxchange > 0.00001
            ' discard last digit
            value = Math.Truncate(value * 10000) / 10000
            ' cleanup
            t2 = DateTime.Now
            Cursor = Cursors.[Default]
            If checkBoxFunctionInfo.Checked Then
                cnt2 = ZetaFunctionInfo.cnt
            Else
                cnt2 = ZetaFunction.cnt
            End If
            sheet.Cells(row, 0).Value = value
            sheet.Columns(0).Width = sheet.GetPreferredColumnWidth(0)
            MessageBox.Show("Zero Value: " + value.ToString() + vbCrLf + "Calculation completed in " + (t2 - t1).ToString() + " with " + (cnt2 - cnt).ToString() + " calls to ZETA function.")
        End Sub
    End Class
End Namespace
